import pandas as pd
import numpy as np
import xgboost as xgb
from sklearn.linear_model import LogisticRegression
import matplotlib.pyplot as plt
import seaborn as sns
import shap
from sklearn.model_selection import train_test_split, TimeSeriesSplit
from sklearn.metrics import roc_auc_score, roc_curve, brier_score_loss, average_precision_score
from skopt import BayesSearchCV
from skopt.space import Real, Integer
from skopt.callbacks import DeadlineStopper, DeltaYStopper
import statsmodels.api as sm


def data_model(df):
    df = df.sort_values(by=['Date'])
    df = df.reset_index(drop=True)

    # Separate the target variable from the features
    X_xgb = df.drop(['Churned_6_Lag','Date','PD_LTV'], axis=1)
    y_xgb = df[['Churned_6_Lag','Missing']]

    X_lr = df.drop(['Churned_6_Lag','Date','PD_LTV'], axis=1)
    y_lr = df[['Churned_6_Lag','Missing']]

    
    # Creating folds for train and test based on date
    train_fold_dates = [
        ('2012-01-01', '2014-12-31'),
        ('2013-01-01', '2015-12-31'),
        ('2014-01-01', '2016-12-31'),
        ('2015-01-01', '2017-12-31'),
        ('2016-01-01', '2018-12-31'),
        ('2017-01-01', '2019-12-31'),
        ('2018-01-01', '2020-12-31'),
        ('2019-01-01', '2021-12-31')
    ]
    test_fold_dates = [
        ('2015-01-01', '2015-12-31'),
        ('2016-01-01', '2016-12-31'),
        ('2017-01-01', '2017-12-31'),
        ('2018-01-01', '2018-12-31'),
        ('2019-01-01', '2019-12-31'),
        ('2020-01-01', '2020-12-31'),
        ('2021-01-01', '2021-12-31'),
        ('2022-01-01', '2022-12-31')
    ]

    tscv = TimeSeriesSplit(n_splits=len(train_fold_dates))
    folds = []
    for i, (train_date_range, test_date_range) in enumerate(zip(train_fold_dates, test_fold_dates)):
        train_start_date, train_end_date = train_date_range
        test_start_date, test_end_date = test_date_range
        
        # Extracting rows corresponding to the current training fold
        train_mask = (df['Date'] >= train_start_date) & (df['Date'] <= train_end_date)
        train_index = np.where(train_mask)[0]
        
        # Extracting rows corresponding to the current test fold
        test_mask = (df['Date'] >= test_start_date) & (df['Date'] <= test_end_date)
        test_index = np.where(test_mask)[0]
        
        folds.append((train_index, test_index))

    

    # Handle missing values separate in train and test by customer
    def missing_values(data):
        grouped = data.groupby('Customer_ID')

        #PD: forward and backward fill
        data['PD'] = grouped['PD'].transform(lambda x: x.fillna(method='ffill'))
        data['PD'] = grouped['PD'].transform(lambda x: x.fillna(method='bfill'))

        # Income: mean + log-transform
        data['Income'] = grouped['Income'].transform(lambda x: x.fillna(x.mean()))
        data['Income_Log'] = np.log(data['Income'])
        data['Income_Log'] = data['Income_Log'].replace([np.inf, -np.inf], 0)

        #Size_Household: mean
        data['Size_Household'] = grouped['Size_Household'].transform(lambda x: x.fillna(x.mean()))

        #LTV: forward and backward fill
        data['LTV'] = grouped['LTV'].transform(lambda x: x.fillna(method='ffill'))
        data['LTV'] = grouped['LTV'].transform(lambda x: x.fillna(method='bfill'))
        return data




    # Building model
    columns = df.columns.drop('Churned_6_Lag')
    list_shap_values = []
    list_test_sets = []
    xgb_scores = []
    lr_scores = []

    def lr_coeff_stats(X_train, y_train, model, fold):
        all_vars = X_train.columns
        result_df = pd.DataFrame()
        
        selected_vars = X_train.columns[model.coef_[0] != 0]
        X_train_selected = X_train[selected_vars]
        X_train_selected = sm.add_constant(X_train_selected)
        logit_model = sm.Logit(y_train, X_train_selected)
        result = logit_model.fit()
        
        coeff_df = pd.DataFrame({'feature': ['const'] + list(X_train_selected.columns[1:]), 'Fold' + str(fold): result.params.round(2).astype(str) + '(' + result.tvalues.round(2).astype(str) + ')'})
        result_df = pd.concat([result_df, coeff_df], axis=0, ignore_index=True)
        print(result_df)
        result_df.to_csv('log_coef'+str(fold)+'.csv', index=False)
        return result_df

    
    def BO_search(X_train, y_train, X_test, y_test):
        # Search space for tuning
        search_spaces = {
            'learning_rate': Real(0.05,0.3),
            'subsample': Real(0.5, 1),
            'gamma': Real(0,1),
            'n_estimators': Integer(30,100),
            'max_depth': Integer(2,6),
            'min_child_weight': Integer(1, 10),
            'reg_lambda': Integer(1,5)
            }
        
        model_xgb_tune = xgb.XGBClassifier(objective='binary:logistic',
                                    eval_metric='logloss',
                                    random_state=1)
        
        gain_control = DeltaYStopper(delta = 0.0001)
        time_control = DeadlineStopper(total_time=3600)

        opt = BayesSearchCV(estimator=model_xgb_tune,
                            search_spaces=search_spaces,
                            scoring = 'roc_auc',
                            n_iter=50,
                            cv = 3)



        opt.fit(X_train, y_train, callback=[gain_control, time_control])
        print(opt.best_params_)
        
        model_xgb = xgb.XGBClassifier(objective='binary:logistic',
                                    eval_metric='logloss',
                                    random_state = 1,
                                    learning_rate=opt.best_params_['learning_rate'],
                                    subsample=opt.best_params_['subsample'],
                                    gamma=opt.best_params_['gamma'],
                                    n_estimators=opt.best_params_['n_estimators'],
                                    max_depth=opt.best_params_['max_depth'],
                                    min_child_weight=opt.best_params_['min_child_weight'],
                                    reg_lambda=opt.best_params_['reg_lambda'])
        return model_xgb 



    model_lr = LogisticRegression(C=0.05, penalty='l1', solver='liblinear', max_iter=1000)



    def model_cv(X,y,model):
        fold = 0

        for train_index, test_index in folds:
            X_train, X_test = X.iloc[train_index], X.iloc[test_index]
            y_train, y_test = y.iloc[train_index], y.iloc[test_index]
            fold += 1
        
            X_train = missing_values(X_train)
            X_test = missing_values(X_test)

            # Drop rest of missing values for X, y
            y_train = y_train.dropna()  
            y_test = y_test.dropna()
            X_train = X_train.dropna()
            X_test = X_test.dropna()


            # Remove "Customer_ID", "Income", "Missing"
            X_train = X_train.drop(['Customer_ID','Income','Missing'], axis=1)
            X_test = X_test.drop(['Customer_ID','Income','Missing'], axis=1)
            y_train = y_train.drop(['Missing'], axis=1)
            y_test = y_test.drop(['Missing'], axis=1)



            if model == "xgb":
                model_xgb = BO_search(X_train, y_train, X_test, y_test) 
                model_xgb.fit(X_train, y_train)
                train_prob = model_xgb.predict_proba(X_train)[:, 1]
                test_prob = model_xgb.predict_proba(X_test)[:, 1]
            else:
                model_lr.fit(X_train, y_train)
                train_prob = model_lr.predict_proba(X_train)[:, 1]
                test_prob = model_lr.predict_proba(X_test)[:, 1]
                lr_coeff_stats(X_train, y_train, model_lr, fold)

            
            train_auc = roc_auc_score(y_train, train_prob)
            train_ap = average_precision_score(y_train, train_prob)
            train_brier = brier_score_loss(y_train, train_prob)
            
            test_auc = roc_auc_score(y_test, test_prob)
            test_ap = average_precision_score(y_test, test_prob)
            test_brier = brier_score_loss(y_test, test_prob)

            if model == "xgb":
                xgb_scores.append({'fold': fold, 'train_auc': train_auc, 'test_auc': test_auc, 'train_ap': train_ap, 'test_ap': test_ap, 'train_brier': train_brier, 'test_brier': test_brier})
                explainer = shap.TreeExplainer(model_xgb)
                sv = explainer(X_train)
                shap_values = sv.values
                list_shap_values.append(shap_values)
                fig = shap.plots.beeswarm(sv, show=False)
                plt.savefig('beeswarm_X_train_fold'+str(fold)+'.png', dpi=600, bbox_inches='tight')
                plt.close()
              
                
            else:
                lr_scores.append({'fold': fold, 'train_auc': train_auc, 'test_auc': test_auc, 'train_ap': train_ap, 'test_ap': test_ap, 'train_brier': train_brier, 'test_brier': test_brier})
            




    # Run models
    model_cv(X_xgb, y_xgb, "xgb")
    #model_cv(X_lr, y_lr, "lr")

    
    def model_best_customer(df):
        best_customer_df = df[df['PD_LTV'] != 0]
        best_customer_df = missing_values(best_customer_df)
        best_customer_df = best_customer_df.dropna()
        best_customer_df = best_customer_df.sort_values(by=['PD_LTV'], ascending=True)
        n = int(len(best_customer_df) * 0.1)
        # Retrieve the first 10% of the rows
        best_customer_df = best_customer_df.head(n)
        X_xgb_best_customer = best_customer_df.drop(['Churned_6_Lag'], axis=1)
        y_xgb_best_customer = best_customer_df[['Churned_6_Lag']]
        y_train = y_xgb_best_customer
        X_train = X_xgb_best_customer.drop(['Customer_ID','Income','Missing','Date'], axis=1)

        # Define model for classification
        model = xgb.XGBClassifier(objective='binary:logistic',
                            eval_metric='logloss',
                            max_depth = 2,
                            gamma = 0.0,
                            reg_lambda = 3,
                            n_estimators = 100,
                            learning_rate = 0.3) 


        # Train the model on the training data
        model.fit(X_train, y_train)

        # Create a SHAP explainer from the trained model
        explainer = shap.Explainer(model)

        # Calculate the SHAP values for each feature in the train data
        shap_values = explainer(X_train)

        # Plot SHAP beeswarm
        fig = shap.plots.beeswarm(shap_values, show=False)
        plt.savefig('beeswarm_best_customer_train.png', dpi=600, bbox_inches='tight')
        plt.close()

        # Plot SHAP waterfall for local interepretation of 28 year old churned customer
        fig = shap.plots.waterfall(shap_values[4201], show=False)
        plt.savefig('waterfall_28.png', dpi=600, bbox_inches='tight')
        plt.close()

        # Plot SHAP waterfall for local interepretation of 58 year old churned customer
        fig = shap.plots.waterfall(shap_values[49096], show=False)
        plt.savefig('waterfall_58.png', dpi=600, bbox_inches='tight')
        plt.close()

        
    model_best_customer(df)

    # Add shapvalues for each fold to dataframe
    columns = df.columns.drop(['Churned_6_Lag','Income','Customer_ID','Missing','Date','PD_LTV'])
    shap_values = np.vstack([sv[1] for sv in list_shap_values])
    sv = np.abs(shap_values)
    mean_sv = np.mean(sv, axis=0)

    shap_df = pd.DataFrame({
        "Feature": columns,
        "Fold1": sv[0],
        "Fold2": sv[1],
        "Fold3": sv[2],
        "Fold4": sv[3],
        "Fold5": sv[4],
        "Fold6": sv[5],
        "Fold7": sv[6],
        "Fold8": sv[7],
        "Mean": mean_sv
    })

    # Extract shap values
    shap_df.to_csv('shap_values.csv', index=False)


    # Add scores from both models to dataframe
    xgb_df = pd.DataFrame(xgb_scores)
    xgb_df.to_csv('xgb_scores.csv', index=False)

    lr_df = pd.DataFrame(lr_scores)
    lr_df.to_csv('lr_scores.csv', index=False)





